from __future__ import annotations

import logging
from pathlib import Path
from typing import TYPE_CHECKING, Literal

from pydantic import BaseModel, ConfigDict, Field, PrivateAttr
from tokenizers import Tokenizer
from transformers import AutoTokenizer, PreTrainedTokenizerFast

from skeletoken.addedtoken import AddedTokens
from skeletoken.decase.decase import decase_vocabulary
from skeletoken.decoders import DecoderDiscriminator
from skeletoken.models import MODELS_THAT_NEED_UNK, ModelDiscriminator, WordPiece, get_subword_prefix_token
from skeletoken.normalizers import LowercaseNormalizer, NormalizerDiscriminator, NormalizerSequence
from skeletoken.padding import Padding
from skeletoken.postprocessors import (
    PostProcessorDiscriminator,
    PostProcessorSequence,
    get_bos_token_from_post_processor,
    get_eos_token_from_post_processor,
    maybe_replace_token_in_post_processor,
)
from skeletoken.pretokenizers import (
    FixedLengthPreTokenizer,
    PreTokenizerDiscriminator,
    PreTokenizerSequence,
    get_metaspace,
)
from skeletoken.truncation import Truncation

if TYPE_CHECKING:
    from skeletoken.model_delta import ModelDelta  # pragma: nocover

logger = logging.getLogger(__name__)


class TokenizerModel(BaseModel):
    model_config = ConfigDict(extra="allow")

    version: Literal["1.0"] = "1.0"
    truncation: None | Truncation = None
    padding: None | Padding = None
    added_tokens: AddedTokens = Field(default_factory=lambda: AddedTokens.model_validate([]))
    normalizer: None | NormalizerDiscriminator = None
    pre_tokenizer: None | PreTokenizerDiscriminator = None
    post_processor: None | PostProcessorDiscriminator = None
    decoder: None | DecoderDiscriminator = None
    model: ModelDiscriminator
    _original_tokenizer: TokenizerModel = PrivateAttr(init=False)

    def model_post_init(self, __context: dict) -> None:
        """Post-initialization processing."""
        self._original_tokenizer = self.model_copy(deep=True)
        # Add any missing added tokens to the vocabulary.
        # Sort to fill up the vocabulary in order.
        for token in sorted(self.added_tokens.root, key=lambda x: x.id):
            # Get the content of the token
            content = token.content
            if content not in self.model.vocab.vocabulary:
                logger.warning(f"Adding missing added token '{content}' to the vocabulary.")
                self._add_token_to_vocabulary(content, is_added_token=True)
            # Retro-actively set the ID.
            curr_id = token.id
            new_id = self.model.vocab[content]
            if curr_id != new_id:
                logger.warning(
                    f"Updating ID of added token '{content}' from {curr_id} to {new_id} to match vocabulary index."
                )
            token.id = new_id
        unk_token = self.unk_token
        if unk_token:
            if unk_token not in self.model.vocab:
                logger.warning(f"Adding unk_token '{unk_token}' to the vocabulary.")
                self._add_token_to_vocabulary(unk_token, is_added_token=True)
            added_token = self.added_tokens.get_token(unk_token)
            if not added_token:
                logger.warning(f"Turning unk_token '{unk_token}' into an AddedToken.")
                self.turn_into_addedtoken(
                    unk_token, is_special=True, normalized=False, single_word=True, lstrip=False, rstrip=False
                )

    def add_addedtoken(
        self,
        token: str,
        is_special: bool = False,
        normalized: bool = False,
        single_word: bool = True,
        lstrip: bool = True,
        rstrip: bool = True,
    ) -> None:
        """Adds an added token to the tokenizer model."""
        self._add_token_to_vocabulary(token, is_added_token=True)
        self.turn_into_addedtoken(
            token, is_special=is_special, normalized=normalized, single_word=single_word, lstrip=lstrip, rstrip=rstrip
        )

    def turn_into_addedtoken(
        self,
        token: str,
        is_special: bool = False,
        normalized: bool = False,
        single_word: bool = True,
        lstrip: bool = True,
        rstrip: bool = True,
    ) -> None:
        """Turns an existing token into an an added token and add it to added_tokens."""
        if token not in self.model.vocab.vocabulary:
            raise ValueError(
                f"Token '{token}' not found in the vocabulary. Please add it first using `add_token_to_vocabulary` or `add_addedtoken`."
            )
        self.added_tokens.maybe_add_token(
            token,
            is_special=is_special,
            normalized=normalized,
            single_word=single_word,
            lstrip=lstrip,
            rstrip=rstrip,
            id=self.model.vocab[token],
        )

    def add_token_to_vocabulary(self, token: str) -> None:
        """Adds a token to the tokenizer's vocabulary."""
        self._add_token_to_vocabulary(token, is_added_token=False)

    def _add_token_to_vocabulary(self, token: str, is_added_token: bool = False) -> None:
        """Adds an added token to the vocabulary."""
        self.model.add_token(token, is_added_token=is_added_token)

    def replace_token_in_vocabulary(self, old_token: str, new_token: str) -> None:
        """Replaces a token with another one."""
        is_added_token = self.added_tokens.get_token(old_token) is not None
        self._replace_token_in_vocabulary(old_token, new_token, is_added_token=is_added_token)

    def _replace_token_in_vocabulary(self, old_token: str, new_token: str, is_added_token: bool = False) -> None:
        """Replaces a token with another one. It keeps the old index in the vocabulary."""
        self.model.replace_token(old_token, new_token, is_added_token=is_added_token)
        self.added_tokens.maybe_replace_token(old_token, new_token)
        if self.post_processor is not None:
            self.post_processor = maybe_replace_token_in_post_processor(
                old_token, new_token, self.model.vocab[new_token], self.post_processor
            )

    def remove_token_from_vocabulary(self, token: str) -> None:
        """Removes a token from the vocabulary."""
        self.model.remove_token(token)
        self.added_tokens.maybe_remove_token(token)

    def decase_vocabulary(self, remove_collisions: bool = False) -> TokenizerModel:
        """
        Decases the vocabulary.

        Parameters
        ----------
        remove_collisions: bool
            If this is set, any tokens that are colliding after lowercasing will be removed from the vocabulary.
            This leads to smaller models, but also a mismatch between tokenizers, which
            should be remedied manually.

        Returns
        -------
        TokenizerModel
            The tokenizer model with a decased vocabulary.

        """
        # Special tokens and unnormalized added tokens need to be skipped.
        special_tokens = self.added_tokens.get_special_tokens() + self.added_tokens.get_unnormalized_tokens()
        sorted_vocab = self.model.vocab.sorted_vocabulary
        vocabulary = decase_vocabulary(
            sorted_vocab,
            [x.content for x in special_tokens],
            is_byte=self.transforms_into_bytes,
            remove_collisions=remove_collisions,
        )
        self.model.vocab.replace_vocabulary(vocabulary)
        if not self.lowercases_input:
            self.add_normalizer(LowercaseNormalizer(), prefix=True)

        return self

    def add_pre_tokenizer(self, pre_tokenizer: PreTokenizerDiscriminator) -> TokenizerModel:
        """Add a pre-tokenizer to the tokenizer model."""
        if self.pre_tokenizer is None:
            self.pre_tokenizer = pre_tokenizer
        elif isinstance(self.pre_tokenizer, PreTokenizerSequence):
            self.pre_tokenizer.pretokenizers.append(pre_tokenizer)
        else:
            self.pre_tokenizer = PreTokenizerSequence(pretokenizers=[self.pre_tokenizer, pre_tokenizer])

        return self

    def add_post_processor(self, post_processor: PostProcessorDiscriminator) -> TokenizerModel:
        """Add a post-processor to the tokenizer model."""
        if self.post_processor is None:
            self.post_processor = post_processor
        elif isinstance(self.post_processor, PostProcessorSequence):
            self.post_processor.processors.append(post_processor)
        else:
            self.post_processor = PostProcessorSequence(processors=[self.post_processor, post_processor])

        return self

    def add_normalizer(self, normalizer: NormalizerDiscriminator, prefix: bool = False) -> TokenizerModel:
        """
        Add a normalizer to the tokenizer model.

        Parameters
        ----------
        normalizer: NormalizerDiscriminator
            The normalizer to add.
        prefix: bool
            Whether to add the normalizer before the other normalizers.
            This can be useful if, for example, one of your normalizers performs
            a destructive transform.

        Returns
        -------
        TokenizerModel
            The tokenizer model with the added normalizer.

        """
        if self.normalizer is None:
            self.normalizer = normalizer
        elif isinstance(self.normalizer, NormalizerSequence):
            self.normalizer.normalizers.append(normalizer)
        else:
            if prefix:
                self.normalizer = NormalizerSequence(normalizers=[normalizer, self.normalizer])
            else:
                self.normalizer = NormalizerSequence(normalizers=[self.normalizer, normalizer])

        return self

    @classmethod
    def from_tokenizer(cls: type[TokenizerModel], tokenizer: Tokenizer) -> TokenizerModel:
        """Create a TokenizerModel from a Tokenizer instance."""
        return cls.from_string(tokenizer.to_str())

    @classmethod
    def from_pretrained(cls: type[TokenizerModel], path_or_repo: str | Path) -> TokenizerModel:
        """Create a TokenizerModel from a pretrained tokenizer."""
        path = Path(path_or_repo)
        if path.exists() and path.is_dir():
            try:
                return cls.from_transformers(str(path))
            except Exception:
                pass
            if not (path / "tokenizer.json").exists():
                raise FileNotFoundError(
                    f"No tokenizer.json found in the directory: {path}. Please provide a valid path."
                )
            # If a tokenizer.json file exists, load it directly
            return cls.from_tokenizer(Tokenizer.from_file(str(path / "tokenizer.json")))
        elif path.exists() and path.is_file():
            return cls.from_tokenizer(Tokenizer.from_file(str(path)))

        return cls._load_remote(path_or_repo)  # pragma: nocover

    @classmethod
    def _load_remote(cls: type[TokenizerModel], path_or_repo: str | Path) -> TokenizerModel:  # pragma: nocover
        """Load a remote tokenizer from a HuggingFace repo."""
        try:
            return cls.from_transformers(str(path_or_repo))
        except Exception as e:
            logger.info(f"Tried to load tokenizer as a Hugging Face tokenizer, but failed: {e}")
            pass
        tokenizer = Tokenizer.from_pretrained(str(path_or_repo))

        return cls.from_tokenizer(tokenizer)

    @classmethod
    def from_string(cls: type[TokenizerModel], json_str: str) -> TokenizerModel:
        """Create a TokenizerModel from a JSON string."""
        return cls.model_validate_json(json_str)

    def to_tokenizer(self) -> Tokenizer:
        """Convert the TokenizerModel back to a Tokenizer instance."""
        return Tokenizer.from_str(self.model_dump_json())

    def make_model_greedy(self, max_input_chars_per_word: int = 100) -> TokenizerModel:
        """Convert the TokenizerModel to a greedy tokenizer model."""
        self.model = self.model.to_greedy()
        assert isinstance(self.model, WordPiece)
        self.model.max_input_chars_per_word = max_input_chars_per_word
        self.add_pre_tokenizer(FixedLengthPreTokenizer(length=max_input_chars_per_word))
        return self

    @property
    def lowercases_input(self) -> bool:
        """Check if the tokenizer lowercases the input."""
        return self.normalizer is not None and self.normalizer.lowercases

    @property
    def transforms_into_bytes(self) -> bool:
        """
        Check if the tokenizer transforms the input into bytes.

        There's two ways this can happen:
            1. If the pretokenizer is a ByteLevelPreTokenizer.
            2. If the normalizer is a ByteLevelNormalizer.

        This is a bit more complicated, because the pretokenizer can be a sequence of pretokenizers,
        and the normalizer can also be a sequence of normalizers.
        """
        if self.pre_tokenizer is not None and self.pre_tokenizer._byte_pretokenizes:
            return True
        if self.normalizer is not None and self.normalizer.byte_normalizes:
            return True
        return False

    @property
    def eos(self) -> list[str] | None:
        """Get the end-of-sequence tokens."""
        if self.post_processor is None:
            return None
        return get_eos_token_from_post_processor(self.post_processor)

    @property
    def bos(self) -> list[str] | None:
        """Get the beginning-of-sequence tokens."""
        if self.post_processor is None:
            return None
        return get_bos_token_from_post_processor(self.post_processor)

    @property
    def splits(self) -> bool:
        """Whether the tokenizer can split the string somehow."""
        if self.pre_tokenizer is None:
            return False
        return self.pre_tokenizer._splits

    @property
    def subword_prefix(self) -> str | None:
        """Get the prefix token, if any."""
        return get_subword_prefix_token(self.model)

    @property
    def word_prefix(self) -> str | None:
        """Get the word prefix token, if any."""
        if self.pre_tokenizer is None:
            return None
        # Word prefixes are not handled by the model, but added
        # by pretokenizers
        if self.transforms_into_bytes:
            return "Ġ"
        return get_metaspace(self.pre_tokenizer)

    @property
    def unk_token(self) -> str | None:
        """Get the unk token, if any."""
        return self.model.unk_token

    @unk_token.setter
    def unk_token(self, token: str | None) -> None:
        """Set the unk token of the tokenizer model."""
        old_unk_token = self.unk_token
        if old_unk_token == token:
            return
        if token is None:
            if isinstance(self.model, MODELS_THAT_NEED_UNK):
                raise ValueError("Cannot unset unk_token for WordPiece or WordLevel models.")
            if old_unk_token is not None:
                logger.info(f"Removing unk_token '{self.model.unk_token}' from the tokenizer.")
            self.model.unk_token = None
            return
        if old_unk_token is None:
            if token not in self.model.vocab:
                logger.info(f"Adding {token} to the vocabulary.")
                self._add_token_to_vocabulary(token, is_added_token=True)
            logger.info(f"Setting unk_token to '{token}'.")
            index = self.model.vocab[token]
            self.added_tokens.maybe_add_token(
                token=token, is_special=True, normalized=True, single_word=True, rstrip=True, lstrip=True, id=index
            )
        elif old_unk_token in self.model.vocab:
            logger.info(f"Changing unk_token from '{old_unk_token}' to '{token}'.")
            self.added_tokens.maybe_replace_token(old_unk_token, token)
            if token not in self.model.vocab:
                self._add_token_to_vocabulary(token, is_added_token=True)
            if self.post_processor is not None:
                self.post_processor = maybe_replace_token_in_post_processor(
                    old_unk_token, token, self.model.vocab[token], self.post_processor
                )
        self.model.unk_token = token

    @property
    def pad_token(self) -> str | None:
        """Get the padding token, if any."""
        if self.padding is None:
            return None
        return self.padding.pad_token

    @pad_token.setter
    def pad_token(self, token: str | None) -> None:
        """Set the padding token of the tokenizer model."""
        if token is None:
            current_token = self.pad_token
            if current_token is not None:
                logger.info("Removing padding token from the tokenizer.")

            self.padding = None
            return
        # No padding module set
        if self.padding is None:
            logger.info(f"Setting padding token to '{token}' and creating a padding module.")
            if token not in self.model.vocab:
                self.model.vocab.add_token(token)
            index = self.model.vocab[token]
            self.padding = Padding(pad_id=index, pad_type_id=0, pad_token=token)
        elif token in self.model.vocab:
            logger.info(f"Changing padding token to existing token '{token}'.")
            self.padding.pad_id = self.model.vocab[token]
            self.padding.pad_token = token
        else:
            logger.info(f"Changing padding token to new token '{token}'.")
            old_pad_token = self.padding.pad_token
            self._replace_token_in_vocabulary(old_pad_token, token, is_added_token=True)
            self.padding.pad_token = token

        # We know token is in vocab here.
        self.added_tokens.maybe_add_token(
            token,
            id=self.model.vocab[token],
            is_special=True,
            normalized=True,
            single_word=True,
            lstrip=True,
            rstrip=True,
        )

    @classmethod
    def from_transformers_tokenizer(cls: type[TokenizerModel], hf_tokenizer: PreTrainedTokenizerFast) -> TokenizerModel:
        """Load a HuggingFace tokenizer from a local path or a model repo."""
        special_tokens = hf_tokenizer.special_tokens_map
        unk_token = special_tokens.get("unk_token", None)
        pad_token = special_tokens.get("pad_token", None)

        model = cls.from_tokenizer(hf_tokenizer._tokenizer)
        if unk_token is not None and isinstance(unk_token, str):
            if model.unk_token is not None and model.unk_token != unk_token:
                logger.warning(
                    f"Overriding existing unk_token '{model.unk_token}' with the one from the HuggingFace tokenizer: '{unk_token}'."
                )
            if model.unk_token is None:
                logger.warning(
                    f"HuggingFace tokenizer defines an unk_token, but the Skeletoken model does not. Setting it to '{unk_token}'."
                )
            model.unk_token = unk_token
        if pad_token is not None and isinstance(pad_token, str):
            if model.pad_token is not None and model.pad_token != pad_token:
                logger.warning(
                    f"Overriding existing pad_token '{model.pad_token}' with the one from the HuggingFace tokenizer: '{pad_token}'."
                )
            if model.pad_token is None:
                logger.warning(
                    f"HuggingFace tokenizer defines a pad_token, but the Skeletoken model does not. Setting it to '{pad_token}'."
                )
            model.pad_token = pad_token

        return model

    @classmethod
    def from_transformers(cls, path: str | Path) -> TokenizerModel:  # pragma: nocover
        """Load a HuggingFace tokenizer from a local path or a model repo."""
        hf_tokenizer: PreTrainedTokenizerFast = AutoTokenizer.from_pretrained(path)
        return cls.from_transformers_tokenizer(hf_tokenizer)

    def to_transformers(
        self, tokenizer_class: type[PreTrainedTokenizerFast] = PreTrainedTokenizerFast
    ) -> PreTrainedTokenizerFast:
        """Convert the TokenizerModel to a HuggingFace tokenizer."""
        tok = tokenizer_class(tokenizer_object=self.to_tokenizer())
        tok.pad_token = self.pad_token
        tok.unk_token = self.unk_token
        if self.bos:
            if len(self.bos) > 1:
                logger.warning(f"Tokenizer has multiple bos tokens: {self.bos}. Not setting it automatically.")
            else:
                tok.bos_token = self.bos[0]
        if self.eos:
            if len(self.eos) > 1:
                logger.warning(f"Tokenizer has multiple eos tokens: {self.eos}. Not setting it automatically.")
            else:
                tok.eos_token = self.eos[0]

        return tok

    @property
    def model_delta(self) -> ModelDelta:
        """Get the model delta between the original tokenizer and the current one."""
        from skeletoken.model_delta import compute_model_delta

        return compute_model_delta(self._original_tokenizer, self)

    def tokens_to_ids(self, tokens: list[str]) -> list[int]:
        """Convert a list of tokens to their corresponding IDs."""
        return [self.model.vocab[token] for token in tokens]

    def ids_to_tokens(self, ids: list[int]) -> list[str]:
        """Convert a list of IDs to their corresponding tokens."""
        inv_vocab = self.model.vocab.sorted_vocabulary
        return [inv_vocab[id] for id in ids]



from math import log
from typing import Any, Literal, overload

import pytest

from skeletoken.base import TokenizerModel
from skeletoken.merges import Merges
from skeletoken.models import BPE, Model, ModelType, Unigram, WordLevel, WordPiece, get_subword_prefix_token
from skeletoken.vocabulary import UnigramVocabulary, Vocabulary


@overload
def _get_default_model(model_type: Literal[ModelType.UNIGRAM]) -> Unigram: ...


@overload
def _get_default_model(model_type: Literal[ModelType.BPE]) -> BPE: ...


@overload
def _get_default_model(model_type: Literal[ModelType.WORDLEVEL]) -> WordLevel: ...


@overload
def _get_default_model(model_type: Literal[ModelType.WORDPIECE]) -> WordPiece: ...


@overload
def _get_default_model(model_type: ModelType) -> Model: ...


def _get_default_model(model_type: ModelType) -> Model:
    """Helper function to get the default instantiation of a model."""
    vocab = {
        "[PAD]": 0,
        "[SEP]": 1,
        "[UNK]": 2,
        "[CLS]": 3,
        "[MASK]": 4,
        "a": 5,
        "b": 6,
        "c": 7,
        "d": 8,
        "e": 9,
        " ": 10,
    }
    if model_type == ModelType.BPE:
        return BPE(
            vocab=Vocabulary(vocab),
            merges=Merges([]),
            dropout=0.1,
            unk_token="[UNK]",
            continuing_subword_prefix="",
            end_of_word_suffix="",
            fuse_unk=False,
            byte_fallback=False,
            ignore_merges=False,
        )
    elif model_type == ModelType.WORDPIECE:
        return WordPiece(
            vocab=Vocabulary(vocab), unk_token="[UNK]", continuing_subword_prefix="", max_input_chars_per_word=100
        )
    elif model_type == ModelType.UNIGRAM:
        p = log(1.0 / len(vocab))
        u_vocab = [(x, p) for x, _ in sorted(vocab.items(), key=lambda item: item[1])]
        return Unigram(vocab=UnigramVocabulary(u_vocab), unk_id=2, byte_fallback=False)
    elif model_type == ModelType.WORDLEVEL:
        return WordLevel(vocab=Vocabulary(vocab), unk_token="[UNK]")
    else:
        raise ValueError(f"Unknown model type: {model_type}")


@pytest.mark.parametrize("model_type", [ModelType.BPE, ModelType.WORDPIECE, ModelType.UNIGRAM, ModelType.WORDLEVEL])
def test_model(small_tokenizer_json: dict[str, Any], model_type: ModelType) -> None:
    """
    Test that the small tokenizer JSON can be loaded and contains the expected structure.

    This test checks that the tokenizer JSON has the correct keys and types for its fields.
    """
    internal_model = _get_default_model(model_type)
    model_dict = internal_model.model_dump()
    small_tokenizer_json["model"] = model_dict
    model = TokenizerModel.model_validate(small_tokenizer_json)

    assert model.model is not None
    assert model.model.type == model_type
    assert isinstance(model.model, internal_model.__class__)

    # Implicit test. If this fails, the model is incorrect.
    model.to_tokenizer()


def _get_none_unigram() -> Unigram:
    """A unigram model with UNK_id set to None."""
    model = _get_default_model(ModelType.UNIGRAM)
    model.unk_id = None

    return model


def _get_none_bpe() -> BPE:
    """A BPE model with UNK_token set to None."""
    model = _get_default_model(ModelType.BPE)
    model.unk_token = None

    return model


@pytest.mark.parametrize("model", [*[_get_default_model(x) for x in ModelType], _get_none_unigram(), _get_none_bpe()])
def test_greedy(model: Model) -> None:
    """Tests the greedy behavior."""
    model = model.to_greedy()
    assert model.type == ModelType.WORDPIECE

    tokenizer_model = TokenizerModel(model=model)
    tokenizer = tokenizer_model.to_tokenizer()

    assert tokenizer.encode("a b c d e").tokens == ["a", " ", "b", " ", "c", " ", "d", " ", "e"]


def test_get_subword_prefix_token() -> None:
    """Tests the continuing subword prefix behavior."""
    wordpiece = _get_default_model(ModelType.WORDPIECE)
    wordpiece.continuing_subword_prefix = "##"

    assert get_subword_prefix_token(wordpiece) == "##"

    wordpiece.continuing_subword_prefix = ""
    assert get_subword_prefix_token(wordpiece) == ""

    bpe = _get_default_model(ModelType.BPE)
    bpe.continuing_subword_prefix = "##"

    assert get_subword_prefix_token(bpe) == "##"

    bpe.continuing_subword_prefix = ""
    assert get_subword_prefix_token(bpe) == ""

    unigram = _get_default_model(ModelType.UNIGRAM)
    assert get_subword_prefix_token(unigram) is None

    wordlevel = _get_default_model(ModelType.WORDLEVEL)
    assert get_subword_prefix_token(wordlevel) is None


def test_unk_token_unigram() -> None:
    """Test the unk token in unigram model."""
    model = _get_default_model(ModelType.UNIGRAM)
    assert model.unk_id == 2
    assert model.unk_token == model.vocab.sorted_vocabulary[2]
    model.unk_token = "a"
    assert model.unk_token == "a"
    assert model.unk_id == model.vocab["a"]

    model.unk_token = None
    assert model.unk_id is None
    assert model.unk_token is None


@pytest.mark.parametrize("model", [*[_get_default_model(x) for x in ModelType]])
def test_add_token(model: Model) -> None:
    """Test the add token functionality."""
    model.add_token("new_token", is_added_token=False)
    assert model.vocab["new_token"] == 11
    if isinstance(model, BPE):
        assert model.vocab.sorted_vocabulary == [
            "[PAD]",
            "[SEP]",
            "[UNK]",
            "[CLS]",
            "[MASK]",
            "a",
            "b",
            "c",
            "d",
            "e",
            " ",
            "new_token",
            "_",
            "k",
            "ke",
            "n",
            "ne",
            "new_",
            "new_toke",
            "o",
            "t",
            "to",
            "toke",
            "w",
            "w_",
        ]
        assert model.merges.root == [
            ("n", "e"),
            ("w", "_"),
            ("t", "o"),
            ("k", "e"),
            ("ne", "w_"),
            ("to", "ke"),
            ("new_", "toke"),
            ("new_toke", "n"),
        ]
    else:
        assert model.vocab.sorted_vocabulary == [
            "[PAD]",
            "[SEP]",
            "[UNK]",
            "[CLS]",
            "[MASK]",
            "a",
            "b",
            "c",
            "d",
            "e",
            " ",
            "new_token",
        ]
    with pytest.raises(ValueError):
        model.add_token("new_token")


@pytest.mark.parametrize("model", [*[_get_default_model(x) for x in ModelType]])
def test_replace_token(model: Model) -> None:
    """Test the replace token functionality."""
    model.replace_token("a", "new_token", is_added_token=False)
    if isinstance(model, BPE):
        assert model.vocab.sorted_vocabulary == [
            "[PAD]",
            "[SEP]",
            "[UNK]",
            "[CLS]",
            "[MASK]",
            "new_token",
            "b",
            "c",
            "d",
            "e",
            " ",
            "_",
            "k",
            "ke",
            "n",
            "ne",
            "new_",
            "new_toke",
            "o",
            "t",
            "to",
            "toke",
            "w",
            "w_",
        ]
        assert model.merges.root == [
            ("n", "e"),
            ("w", "_"),
            ("t", "o"),
            ("k", "e"),
            ("ne", "w_"),
            ("to", "ke"),
            ("new_", "toke"),
            ("new_toke", "n"),
        ]
    else:
        assert model.vocab.sorted_vocabulary == [
            "[PAD]",
            "[SEP]",
            "[UNK]",
            "[CLS]",
            "[MASK]",
            "new_token",
            "b",
            "c",
            "d",
            "e",
            " ",
        ]
    with pytest.raises(ValueError):
        model.add_token("new_token")


@pytest.mark.parametrize("model", [*[_get_default_model(x) for x in ModelType]])
def test_remove_token(model: Model) -> None:
    """Test the remove token functionality."""
    model.remove_token("a")
    assert "a" not in model.vocab
    assert model.vocab.sorted_vocabulary == [
        "[PAD]",
        "[SEP]",
        "[UNK]",
        "[CLS]",
        "[MASK]",
        "b",
        "c",
        "d",
        "e",
        " ",
    ]
    with pytest.raises(ValueError):
        model.remove_token("a")

    model.replace_token("b", "x", is_added_token=True)
    # X gets added to the vocabulary, but not as a merge.
    if isinstance(model, BPE):
        assert "x" not in model.merges._all_merge_tokens
